PUB start(SignalPtr)

'' Start FFT Driver - starts a COG
'' returns false if no COG available
''
''    SignalPtr = Pointer to the data
''
''
''
'  SignalPtr point to several arrays in memory
'  1024 16 bit values for real data
'  1024 16 bit values for imaginary data (cleared)
'  
'  This demo takes 1024 samples @ 1 Msps
'
'  AD9832 is connected as follows
'  D0..D7 to PA0..PA7
'  ADCEN  to PA26
'  ADCCLK to PA27
'
'  A VGA monitor connected between PA16 and PA23 as usual is also needed !!
'
  
  cognew(@init, SignalPtr)

CON
NN=1024
BITS_NN=10
BITS_NNM1=9
BITS_DIFF=3
k_ADCEN    = 1<<26
k_ADCCLK   = 1<<27
  
DAT
                        org
'.section cog cog0

' Integer FFT
' 16 bit signed values are used
' it works v4.0
'
' Needs 4096 bytes of HUB RAM



init                    mov     cnt_bitmap_ptr,PAR
                        mov     fft_ii,#16                  ' displays bitmap
                        mov     fft_jj,cnt_k1_512
                        mov     fft_m,cnt_bitmap_ptr

init_2                  wrlong  bitmap0,fft_m
                        add     fft_m,#64
                        add     init_2,fft_jj
                        djnz    fft_ii,#init_2
                        
                        mov     DIRA,cnt_k1_DIRA

                        
' Gets 1024 samples from the adc (AD9832)
do_adc                  andn    OUTA,cnt_k1_ADCEN       ' enables ADC
                        mov     k1_delay,#40            ' 500 ns, so 1msps
                        mov     k1_delay2,#36
                        
c1_prepadc              mov     k1_a,#5 ' 5 dummy adcs
                        mov     k1_ptr,cnt_k1_CH0BUFF
                        mov     k1_wait,CNT
                        add     k1_wait,k1_delay
c1_dummyadc             or      OUTA,cnt_k1_ADCCLK
                        waitcnt k1_wait,k1_delay
                        andn    OUTA,cnt_k1_ADCCLK
                        waitcnt k1_wait,k1_delay2
                        djnz    k1_a,#c1_dummyadc

                        mov     k1_a,#1               ' length
                        shl     k1_a,#BITS_NN              ' 1024 samples
                        sub     k1_ptr,#2
                        waitcnt k1_wait,k1_delay

c1_realadc              or      OUTA,cnt_k1_ADCCLK
                        mov     k1_b,INA
                        shl     k1_b,#24
                        sar     k1_b,#18
                        'and     k1_b,#cnt_k1_mask
                        add     k1_ptr,#2
                        waitcnt k1_wait,k1_delay
                        andn    OUTA,cnt_k1_ADCCLK
                        wrword  k1_b,k1_ptr             ' writes data to real part
                        waitcnt k1_wait,k1_delay2              
                        djnz    k1_a,#c1_realadc        ' gets the ADC done

                        mov     OUTA,cnt_k1_ADCEN       ' disables ADC
' clears imaginary array                        
                        mov     k1_a,cnt_k1_512                        
                        mov     k1_b,#0
c1_clearimag            wrlong  k1_b,k1_ptr             ' clears imaginary part
                        add     k1_ptr,#4
                        djnz    k1_a,#c1_clearimag
                        call    #plot_waveform                        
' FFT starts here
                        mov     fft_fr,cnt_rsample_ptr
                        mov     fft_fi,cnt_isample_ptr

                        mov     fft_n,#1
                        shl     fft_n,#BITS_NN          '1024 point fft
                        call    #decimate
                        call    #lets_rock
                        call    #calc_abs
                        call    #plot
                        'mov     fft_ii,#2
                        'wrlong  fft_ii,cnt_dump_ready
                        jmp     #do_adc                ' new round

decimate                mov     fft_ii,#1
                        mov     fft_ll,fft_n
ldecimate               mov     fft_jj,fft_ii
                        rev     fft_jj,#22 'BITS_NN
                        cmp     fft_ii,fft_jj   wc
              if_nc     jmp     #ldecimate_5
                        mov     fft_fr_ii,fft_ii
                        shl     fft_fr_ii,#1
                        add     fft_fr_ii,fft_fr

                        mov     fft_fr_jj,fft_jj

                        rdword  fft_tr,fft_fr_ii

                        shl     fft_fr_jj,#1
                        add     fft_fr_jj,fft_fr
                        rdword  fft_result,fft_fr_jj

                        wrword  fft_tr,fft_fr_jj
                        wrword  fft_result,fft_fr_ii
ldecimate_5             add     fft_ii,#1
                        cmp     fft_ii,fft_ll   wc
              if_c      jmp     #ldecimate
decimate_ret            ret

' Calcs the 1024 point-FFT using 16 bit signed integers
'
lets_rock               mov     fft_ll,#1
                        mov     fft_k,#BITS_NNM1           ' 9 bits for 512 points

lets_rock_while         cmp     fft_ll,fft_n wc
              if_nc     jmp     #lets_rock_while_e                                          

                        mov     fft_is,fft_ll
                        shl     fft_is,#1
                        mov     fft_m,#0
lets_rock_for_1         cmp     fft_m,fft_ll wc
              if_nc     jmp     #lets_rock_for_1_e

                        mov     fft_jj,fft_m
                        shl     fft_jj,fft_k

                        call    #get_sincos
                        mov     fft_ii,fft_m
lets_rock_for_2         cmp     fft_ii,fft_n wc
              if_nc     jmp     #lets_rock_for_2_e

                        mov     fft_jj,fft_ii
                        add     fft_jj,fft_ll

                        mov     fft_fi_jj,fft_jj
                        shl     fft_fi_jj,#1      ' word access
                        mov     fft_fr_jj,fft_fi_jj
                        add     fft_fr_jj,fft_fr
                        add     fft_fi_jj,fft_fi

                        rdword  fft_result,fft_fr_jj
                        call    #lets_mul_wr
                        mov     fft_tr,fft_result
                        rdword  fft_result,fft_fi_jj
                        call    #lets_mul_wi
                        subs    fft_tr,fft_result     ' 32 bit signed value

                        rdword  fft_result,fft_fi_jj
                        call    #lets_mul_wr
                        mov     fft_ti,fft_result
                        rdword  fft_result,fft_fr_jj
                        call    #lets_mul_wi
                        adds    fft_ti,fft_result     ' 32 bit signed value

                        mov     fft_fi_ii,fft_ii
                        shl     fft_fi_ii,#1      ' word access
                        mov     fft_fr_ii,fft_fi_ii
                        add     fft_fr_ii,fft_fr
                        add     fft_fi_ii,fft_fi
                        
                        rdword  fft_qr,fft_fr_ii     ' qr = fr[i]
                        shl     fft_qr,#16
                        sar     fft_qr,#1         ' scales to 32 bit signed value
                        mov     fft_result,fft_tr
                        rdword  fft_qi,fft_fi_ii     ' qi = fi[i]
                        shl     fft_qi,#16
                        sar     fft_qi,#1         ' scales to 32 bit signed value
                        adds    fft_result,fft_qr     ' res = tr + qr
                        subs    fft_qr,fft_tr         ' qr = qr - tr
                        shr     fft_result,#16     ' scales down
                        wrword  fft_result,fft_fr_ii ' fr[i] = res = tr + qr

                        mov     fft_result,fft_ti
                        adds    fft_result,fft_qi     ' res = ti + qi
                        shr     fft_qr,#16         ' scales down
                        wrword  fft_qr,fft_fr_jj     ' fr[j] = qr = qr - tr
                        subs    fft_qi,fft_ti         ' qi = qi - ti
                        shr     fft_result,#16     ' scales down
                        wrword  fft_result,fft_fi_ii ' fi[i] = ti + qi
                        shr     fft_qi,#16         ' scales down
                        add     fft_ii,fft_is
                        wrword  fft_qi,fft_fi_jj     ' fi[j] = qi = qi - ti     
                        jmp     #lets_rock_for_2
lets_rock_for_2_e       
                        add     fft_m,#1
                        jmp     #lets_rock_for_1
lets_rock_for_1_e       sub     fft_k,#1
                        mov     fft_ll,fft_is
                        jmp     #lets_rock_while
lets_rock_while_e
lets_rock_ret           ret


lets_mul_wi             mov     fft_sgn,fft_result
                        and     fft_sgn,cnt_sgn   wz
                        shl     fft_result,#16
                        negnz   fft_result,fft_result
                        shr     fft_result,#15
                       
                        shr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc
              if_c      add     fft_result,fft_wi wc
                        rcr     fft_result,#1 wc
                        xor     fft_sgn,fft_sgnwi   wz
                        negnz   fft_result,fft_result
lets_mul_wi_ret         ret

lets_mul_wr             mov     fft_sgn,fft_result
                        and     fft_sgn,cnt_sgn   wz
                        shl     fft_result,#16
                        negnz   fft_result,fft_result
                        shr     fft_result,#15
                        shr     fft_result,#1 wc
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_wr wc
                        rcr     fft_result,#1 wc
                        xor     fft_sgn,fft_sgnwr   wz
                        negnz   fft_result,fft_result
lets_mul_wr_ret         ret

' Uses the ROM table to get the sine and cosine of jj

get_sincos              mov     fft_wr,fft_jj
                        shl     fft_wr,#BITS_DIFF
                        mov     fft_wi,fft_wr
                        add     fft_wr,cnt_sin_90
                        test    fft_wi,cnt_sin_90 wc
                        test    fft_wi,cnt_sin_180 wz
                        negc    fft_wi,fft_wi
                        or      fft_wi,cnt_sin_table1
                        shl     fft_wi,#1
                        rdword  fft_wi,fft_wi
              if_z      mov     fft_sgnwi,cnt_sgn        ' they are inverted
              if_nz     mov     fft_sgnwi,#0
                        test    fft_wr,cnt_sin_90 wc
                        test    fft_wr,cnt_sin_180 wz
                        negc    fft_wr,fft_wr
                        or      fft_wr,cnt_sin_table1
                        shl     fft_wr,#1
                        rdword  fft_wr,fft_wr
              if_nz     mov     fft_sgnwr,cnt_sgn        ' they are inverted
              if_z      mov     fft_sgnwr,#0

                        shl     fft_wr,#14
                        shl     fft_wi,#14
get_sincos_ret          ret

lets_mul_qr             mov     fft_result,fft_qr
                        shl     fft_result,#16
                        abs     fft_result,fft_result
                        mov     fft_qr,fft_result
                        shr     fft_result,#16
                        
                        shr     fft_result,#1 wc
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1 wc                        
              if_c      add     fft_result,fft_qr wc
                        rcr     fft_result,#1
lets_mul_qr_ret         ret

lets_sqrt_qi            mov     fft_result,#0
                        mov     fft_m,#0
                        mov     fft_jj,#16
lets_sqrt_qi_l          shl     fft_qi,#1 wc
                        rcl     fft_m,#1
                        shl     fft_qi,#1 wc
                        rcl     fft_m,#1
                        shl     fft_result,#2
                        or      fft_result,#1
                        cmpsub  fft_m,fft_result wc, wr
                        shr     fft_result,#2
                        rcl     fft_result,#1
                        djnz    fft_jj,#lets_sqrt_qi_l
lets_sqrt_qi_ret        ret

calc_abs                mov      fft_ii,cnt_k1_512
                        mov      fft_fr_ii,fft_fr
                        mov      fft_fi_ii,fft_fi
calc_abs_5              rdword   fft_qr,fft_fr_ii
                        call     #lets_mul_qr
                        mov      fft_qi,fft_result
                        rdword   fft_qr,fft_fi_ii
                        add      fft_fi_ii,#2    ' next word
                        call     #lets_mul_qr
                        add      fft_qi,fft_result
                        call     #lets_sqrt_qi
                        shr      fft_result,#4
                        wrword   fft_result,fft_fr_ii
                        add      fft_fr_ii,#2    ' next word
                        djnz     fft_ii,#calc_abs_5
calc_abs_ret            ret

' Plots the spectrum.
' Left most point is 1, rightmost is 128 !!!
'

plot                    mov      fft_ii,#64
                        mov      fft_jj,#0

                        mov      fft_fr_ii,fft_fr
                        
plot_8p                 mov      fft_m,#256
                        mov      fft_k,cnt_bitmap_ptr
                        add      fft_k,cnt_128m64
                        mov      fft_qr,#0
                        add      fft_k,fft_jj
plot_8p_l               wrbyte   fft_qr,fft_k   ' clears bitmap
                        add      fft_k,#64
                        djnz     fft_m,#plot_8p_l                     

                        mov      fft_k,#$01
                        rdword   fft_qr,fft_fr_ii
                        add      fft_fr_ii,#2
                        call     #putpix
                        shl      fft_k,#1
                        rdword   fft_qr,fft_fr_ii
                        add      fft_fr_ii,#2
                        call     #putpix
                        shl      fft_k,#1
                        rdword   fft_qr,fft_fr_ii
                        add      fft_fr_ii,#2
                        call     #putpix
                        shl      fft_k,#1
                        rdword   fft_qr,fft_fr_ii
                        add      fft_fr_ii,#2
                        call     #putpix
                        shl      fft_k,#1
                        rdword   fft_qr,fft_fr_ii
                        add      fft_fr_ii,#2
                        call     #putpix
                        shl      fft_k,#1
                        rdword   fft_qr,fft_fr_ii
                        add      fft_fr_ii,#2
                        call     #putpix
                        shl      fft_k,#1
                        rdword   fft_qr,fft_fr_ii
                        add      fft_fr_ii,#2
                        call     #putpix
                        shl      fft_k,#1
                        rdword   fft_qr,fft_fr_ii
                        add      fft_fr_ii,#2
                        call     #putpix
                        add      fft_jj,#1
                        djnz     fft_ii,#plot_8p

plot_ret                ret

                         

putpix                  mov      fft_qi,#255
                        max      fft_qi,fft_qr
                        mov      fft_qr,#383
                        sub      fft_qr,fft_qi
                        shl      fft_qr,#6
'                        mov      fft_qi,fft_qr
'                        shl      fft_qr,#2
'                        add      fft_qr,fft_qi
                        add      fft_qr,cnt_bitmap_ptr
                        add      fft_qr,fft_jj
                        rdbyte   fft_ll,fft_qr
                        or       fft_ll,fft_k
                        wrbyte   fft_ll,fft_qr
putpix_ret              ret

plot_waveform           mov      fft_fr_ii,fft_fr       ' pointer to data
                        mov      fft_ii,#64             ' 64 grouaps of 8 points

                        mov      fft_jj,#0
' clears wave area                        
plot_w_8p               mov      fft_m,#128
                        mov      fft_k,cnt_bitmap_ptr
                        mov      fft_qr,#0
                        add      fft_k,fft_jj
plot_w_8p_l             wrbyte   fft_qr,fft_k   ' clears bitmap
                        add      fft_k,#64
                        djnz     fft_m,#plot_w_8p_l                        

                        mov      fft_k,#$01
                        'mov      fft_m,
plot_w_1p               rdword   fft_qr,fft_fr_ii
                        shr      fft_qr,#1
                        and      fft_qr,cnt_k1_3fc0
                        add      fft_qr,fft_jj  ' byte index
                        add      fft_qr,cnt_bitmap_ptr
                        rdbyte   fft_qi,fft_qr
                        or       fft_qi,fft_k
                        wrbyte   fft_qi,fft_qr
                        add      fft_fr_ii,#2   ' next data
                        shl      fft_k,#1
                        cmp      fft_k,#$100 wz
              if_nz     jmp      #plot_w_1p
                        add      fft_jj,#1
                        djnz     fft_ii,#plot_w_8p
plot_waveform_ret       ret

' constants
cnt_sgn                 long    $8000
cnt_sin_90              long    $0800
cnt_sin_180             long    $1000
cnt_sin_table           long    0 ' $e000
cnt_sin_table1          long    $7000
cnt_dump_ready          long    $7fe4
cnt_rsample_ptr         long    $0000 '1400
cnt_isample_ptr         long    $800
cnt_bitmap_ptr          long    $2000 '4996 ' $1fc0-48*64+60
cnt_k1_512              long    512

bitmap0                 long    %0001_1111_1111_1111_1111_1000_0000_0000
bitmap1                 long    %0010_0000_0000_0000_0000_0100_0000_0000
bitmap2                 long    %0100_0000_0000_0000_0000_0010_0000_0000
bitmap3                 long    %1000_0000_0000_0000_0000_0001_0000_0000
bitmap4                 long    %1001_1111_0111_1101_1111_0001_0000_0000
bitmap5                 long    %1000_0100_0000_0100_0001_0001_0000_0000
bitmap6                 long    %1000_0100_0000_0100_0001_0001_0000_0000
bitmap7                 long    %1000_0100_0011_1100_1111_0001_0000_0000
bitmap8                 long    %1000_0100_0000_0100_0001_0001_0000_0000
bitmap9                 long    %1000_0100_0000_0100_0001_0001_0000_0000
bitmapa                 long    %1000_0100_0000_0100_0001_0001_0000_0000
bitmapb                 long    %1000_0000_0000_0000_0000_0001_0000_0000
bitmapc                 long    %0100_0000_0000_0000_0000_0010_0000_0000
bitmapd                 long    %0010_0000_0000_0000_0000_0100_0000_0000
bitmape                 long    %0001_1111_1111_1111_1111_1000_0000_0000
bitmapf                 long    %0000_0000_0000_0000_0000_0000_0000_0000

' Variables

fft_ii                  long    0
fft_is                  long    0
fft_jj                  long    0
fft_k                   long    0
fft_ll                  long    0
fft_m                   long    0
fft_n                   long    0

fft_qr                  long    0
fft_qi                  long    0
fft_fr                  long    0
fft_fi                  long    0
fft_tr                  long    0
fft_ti                  long    0
fft_wr                  long    0
fft_wi                  long    0
fft_fi_ii               long    0
fft_fr_ii               long    0
fft_fi_jj               long    0
fft_fr_jj               long    0
fft_result              long    0
fft_sgn                 long    0
fft_sgnwr               long    0  ' sign of wr
fft_sgnwi               long    0  ' sign of wi

k1_ptr        long      0              
k1_a          long      0
k1_b          long      0
k1_c          long      0
k1_wait       long      0
k1_delay      long      0
k1_delay2     long      0

cnt_k1_ADCEN  long      k_ADCEN
cnt_k1_ADCCLK long      k_ADCCLK
cnt_k1_DIRA   long      k_ADCEN|k_ADCCLK
cnt_k1_CH0BUFF long     $0000 ' k_CH0BUFF
cnt_k1_MASK   long      $ffff
cnt_128m64    long      128*64
cnt_k1_3fc0   long      $1fc0
